# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from itertools import islice
import math
from typing import Any, Iterable, TypeVar


from .predictable import Predictable
from ..configs.sampling import get_predictable_numbers  # , predictable_numbers

EOI = object

logger = logging.getLogger(__name__)


T_co = TypeVar("T_co", covariant=True)


def reservoir_sample_r(iterable: Iterable[T_co], k: int, seed: int) -> list[T_co]:
    """
    Selects a random sample of k items from an iterable using Algorithm R.

    Note: Algorithm R is used over L for cross-compatability (i.e. no dependencies of math.log, math.exp etc.)

    Args:
        iterable: The iterable to sample from.
        k: The desired sample size (number of items to select).
        seed: The seed for the `Predictable` number generator

    Returns:
        A list containing the random sample of k items. If the iterable
        contains fewer than k items, all items from the iterable are returned.
    """
    if k < 0:
        raise ValueError("Sample size k cannot be negative.")
    if k == 0:
        return []

    it = iter(iterable)

    reservoir: list[T_co] = list(islice(it, k))

    if len(reservoir) < k:
        return reservoir

    png = Predictable(sequence=get_predictable_numbers(), seed=seed)
    for i, item in enumerate(it, start=k):
        j = png.randrange(0, i + 1)
        if j < k:
            reservoir[j] = item
    return reservoir



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from dorsal.common import constants
from dorsal.common.config import load_config, resolve_setting

logger = logging.getLogger(__name__)


def _get_cache_enabled_from_env() -> bool | None:
    """Reads the DORSAL_CACHE_ENABLED environment variable."""
    env_var = os.getenv(constants.ENV_DORSAL_CACHE_ENABLED)
    if env_var is None:
        return None
    return env_var.lower() not in ("false", "0", "no")


def _get_cache_enabled_from_config() -> bool | None:
    """Reads the 'enabled' flag from the [cache] section of dorsal.toml."""
    config, _ = load_config()
    config_val = config.get(constants.CONFIG_SECTION_CACHE, {}).get(constants.CONFIG_OPTION_ENABLED)

    if not isinstance(config_val, bool):
        if config_val is not None:
            logger.warning("Invalid value '%s' for enabled flag in config. Ignoring.", config_val)
        return None

    return config_val


def get_cache_enabled(use_cache: bool | None = None) -> bool:
    """Resolves whether the cache is enabled with standard precedence."""
    return resolve_setting(
        setting_name="cache_enabled",
        explicit_value=use_cache,
        env_getter=_get_cache_enabled_from_env,
        config_getter=_get_cache_enabled_from_config,
        default_value=True,
    )


def _get_cache_compression_from_env() -> bool | None:
    """Reads the DORSAL_CACHE_COMPRESSION environment variable."""
    env_var = os.getenv(constants.ENV_DORSAL_CACHE_COMPRESSION)
    if env_var is None:
        return None
    return env_var.lower() not in ("false", "0", "no")


def _get_cache_compression_from_config() -> bool | None:
    """Reads the 'compression' flag from the [cache] section of dorsal.toml."""
    config, _ = load_config()
    config_val = config.get(constants.CONFIG_SECTION_CACHE, {}).get(constants.CONFIG_OPTION_COMPRESSION)

    if not isinstance(config_val, bool):
        if config_val is not None:
            logger.warning(
                "Invalid value '%s' for compression flag in config. Ignoring.",
                config_val,
            )
        return None

    return config_val


def get_cache_compression(compress: bool | None = None) -> bool:
    """Resolves whether cache compression is enabled with standard precedence."""
    return resolve_setting(
        setting_name="cache_compression",
        explicit_value=compress,
        env_getter=_get_cache_compression_from_env,
        config_getter=_get_cache_compression_from_config,
        default_value=True,
    )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from typer.testing import CliRunner
from unittest.mock import MagicMock

from dorsal.cli import app

runner = CliRunner()


@pytest.fixture
def mock_clear_cache_cmd(mocker):
    """Mocks dependencies for the `cache clear` command."""
    mock_get_cache = mocker.patch("dorsal.session.get_shared_cache")
    mock_clear_shared_cache = mocker.patch("dorsal.session.clear_shared_cache")

    mock_cache_instance = MagicMock()
    mock_cache_instance.db_path.resolve.return_value = "/fake/home/.dorsal/cache.db"
    mock_get_cache.return_value = mock_cache_instance

    return {
        "get_cache": mock_get_cache,
        "clear_shared_cache": mock_clear_shared_cache,
        "cache_instance": mock_cache_instance,
    }


def test_clear_cache_with_yes_flag(mock_rich_console, mock_clear_cache_cmd):
    """Tests clearing the cache non-interactively with the --yes flag."""
    result = runner.invoke(app, ["cache", "clear", "--yes"])

    assert result.exit_code == 0
    mock_clear_cache_cmd["cache_instance"].clear.assert_called_once()
    mock_clear_cache_cmd["clear_shared_cache"].assert_called_once()

    success_msg = mock_rich_console.print.call_args.args[0]
    assert "Cache cleared successfully" in success_msg


def test_clear_cache_with_confirmation_yes(mock_rich_console, mock_clear_cache_cmd):
    """Tests clearing the cache by answering 'y' to the interactive prompt."""

    result = runner.invoke(app, ["cache", "clear"], input="y\n")

    assert result.exit_code == 0
    mock_clear_cache_cmd["cache_instance"].clear.assert_called_once()
    mock_clear_cache_cmd["clear_shared_cache"].assert_called_once()

    success_msg = mock_rich_console.print.call_args.args[0]
    assert "Cache cleared successfully" in success_msg


def test_clear_cache_cancelled(mock_rich_console, mock_clear_cache_cmd):
    """Tests cancelling the clear command by answering 'n' to the prompt."""
    result = runner.invoke(app, ["cache", "clear"], input="n\n")

    assert result.exit_code == 0
    cancelled_msg = mock_rich_console.print.call_args.args[0]
    assert "Cache clearing cancelled" in cancelled_msg

    mock_clear_cache_cmd["cache_instance"].clear.assert_not_called()
    mock_clear_cache_cmd["clear_shared_cache"].assert_not_called()


def test_clear_cache_exception_handling(mock_clear_cache_cmd):
    """Tests that an exception during the clear operation is handled gracefully."""
    mock_clear_cache_cmd["cache_instance"].clear.side_effect = OSError("Permission denied")

    result = runner.invoke(app, ["cache", "clear", "--yes"])

    assert result.exit_code != 0
    assert "An error occurred while clearing the cache: Permission denied" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from dorsal.file.validators.file_record import FileRecordStrict


def normalize_record_privacy(
    record: FileRecordStrict, target_private: bool | None = None, strict: bool = False
) -> FileRecordStrict:
    """
    Mutates the record in-place to normalize the `private` field of tags/annotations.

    Args:
        record: The record to mutate.
        target_private: The target state (True/False) or None to neutralize.
        strict: If True, raises ValueError on conflicts, else simply modifies the value.

    Raises:
        ValueError: If strict=True and a mismatch is found.
    """
    from dorsal.file.validators.file_record import Annotation, AnnotationGroup, AnnotationsStrict

    def _apply_policy(item: Any, location_desc: str) -> None:
        if not hasattr(item, "private"):
            return

        current_val = item.private

        # Case A: Neutralize (Client Mode) -> Wipe it
        if target_private is None:
            item.private = None
            return

        # Case B: Resolve (Server Mode)
        if current_val is None:
            # Agnostic -> Resolve
            item.private = target_private
        elif current_val != target_private:
            # Conflict detected
            if strict:
                required_state = "PRIVATE" if target_private else "PUBLIC"
                actual_state = "PRIVATE" if current_val else "PUBLIC"
                raise ValueError(
                    f"Privacy Mismatch in {location_desc}: Endpoint requires {required_state} records, "
                    f"but received an item explicitly marked {actual_state}."
                )
            else:
                # Overwrite
                item.private = target_private

    def _process_container(container: Any, path: str) -> None:
        """Recursive traverser for Lists, Groups, and Items."""
        if isinstance(container, AnnotationGroup):
            # Group container itself doesn't have privacy, only children
            for i, ann in enumerate(container.annotations):
                _apply_policy(ann, f"{path} -> Group Index {i}")

        elif isinstance(container, list):
            for i, item in enumerate(container):
                _process_container(item, f"{path}[{i}]")

        elif isinstance(container, Annotation):
            _apply_policy(container, path)

        # Fallback for dynamic/extra fields that might be raw objects
        elif hasattr(container, "private"):
            _apply_policy(container, path)

    # 1. Normalize Tags
    if record.tags:
        for i, tag in enumerate(record.tags):
            _apply_policy(tag, f"Tag index {i} ('{tag.name}')")

    # 2. Normalize Annotations
    if record.annotations:
        # A. Core Fields
        for field_name in AnnotationsStrict.model_fields:
            entry = getattr(record.annotations, field_name, None)
            if entry:
                _process_container(entry, f"Annotation '{field_name}'")

        # B. Extra Fields (Dynamic)
        if record.annotations.__pydantic_extra__:
            for key, value in record.annotations.__pydantic_extra__.items():
                _process_container(value, f"Annotation '{key}'")

    return record



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel, Field

from dorsal.api.adapters import (
    export_record,
    export_record_to_file,
    parse_file,
    parse_file_from_path,
    get_supported_formats,
    get_format_extension,
)
from dorsal.common.exceptions import DorsalError


class DummyRecord(BaseModel):
    """A dummy Pydantic model for testing BaseModel normalization."""

    text_field: str = Field(alias="textField")
    number: int


@patch("dorsal.api.adapters.get_adapter")
def test_export_record_with_dict(mock_get_adapter):
    """Test exporting a standard python dictionary."""
    mock_adapter_instance = MagicMock()
    mock_adapter_instance.export.return_value = "Mocked Export String"
    mock_get_adapter.return_value = mock_adapter_instance

    record = {"foo": "bar"}
    result = export_record(record, schema_id="test/schema", target_format="srt")

    assert result == "Mocked Export String"
    mock_get_adapter.assert_called_once_with("test/schema", "srt")
    mock_adapter_instance.export.assert_called_once_with(record)


@patch("dorsal.api.adapters.get_adapter")
def test_export_record_with_pydantic_model(mock_get_adapter):
    """Test that Pydantic models are correctly normalized to dictionaries before export."""
    mock_adapter_instance = MagicMock()
    mock_adapter_instance.export.return_value = "Mocked Export String"
    mock_get_adapter.return_value = mock_adapter_instance

    record = DummyRecord(textField="Hello World", number=42)
    result = export_record(record, schema_id="test/schema", target_format="vtt")

    assert result == "Mocked Export String"
    expected_dict = {"textField": "Hello World", "number": 42}
    mock_adapter_instance.export.assert_called_once_with(expected_dict)


@patch("dorsal.api.adapters.get_adapter")
def test_export_record_with_kwargs(mock_get_adapter):
    """Test that kwargs are threaded through to the adapter export method."""
    mock_adapter_instance = MagicMock()
    mock_get_adapter.return_value = mock_adapter_instance

    record = {"foo": "bar"}
    export_record(record, schema_id="test/schema", target_format="tsv", include_speakers=True)

    mock_adapter_instance.export.assert_called_once_with(record, include_speakers=True)


@patch("dorsal.api.adapters.get_adapter")
def test_export_record_handles_exceptions(mock_get_adapter):
    """Test that adapter export failures are caught and wrapped in a DorsalError."""
    mock_get_adapter.side_effect = ValueError("Unsupported format 'fake'")

    with pytest.raises(DorsalError, match="Failed to export record to fake: Unsupported format 'fake'"):
        export_record({"key": "value"}, schema_id="test/schema", target_format="fake")


@patch("dorsal.api.adapters.export_record")
def test_export_record_to_file_success(mock_export_record, tmp_path):
    """Test standard file export and format inference from the file extension."""
    mock_export_record.return_value = "Mocked Export String"

    out_file = tmp_path / "output_test.srt"

    result = export_record_to_file({"foo": "bar"}, out_file, "test/schema")

    assert result == "Mocked Export String"
    assert out_file.read_text(encoding="utf-8") == "Mocked Export String"

    mock_export_record.assert_called_once_with({"foo": "bar"}, schema_id="test/schema", target_format="srt")


def test_export_record_to_file_inference_failure(tmp_path):
    """Test error handling when the format cannot be inferred from the output path."""
    out_file = tmp_path / "file_without_extension"

    with pytest.raises(DorsalError, match="Could not infer target format from file extension"):
        export_record_to_file({"foo": "bar"}, out_file, "test/schema")


@patch("dorsal.api.adapters.export_record")
def test_export_record_to_file_write_failure(mock_export_record, tmp_path):
    """Test that OS-level write errors are safely wrapped in a DorsalError."""
    mock_export_record.return_value = "Mocked Export String"
    out_file = tmp_path / "output_test.srt"

    with patch("builtins.open", side_effect=PermissionError("Access Denied")):
        with pytest.raises(DorsalError, match="Failed to write exported text to file"):
            export_record_to_file({"foo": "bar"}, out_file, "test/schema")


@patch("dorsal.api.adapters.get_adapter")
def test_parse_file_with_string_and_kwargs(mock_get_adapter):
    """Test parsing raw string content directly, alongside passing kwargs."""
    mock_adapter_instance = MagicMock()
    mock_adapter_instance.parse.return_value = {"parsed": "data"}
    mock_get_adapter.return_value = mock_adapter_instance

    raw_content = "some raw text"
    result = parse_file(raw_content, schema_id="test/schema", source_format="txt", strict=True)

    assert result == {"parsed": "data"}
    mock_adapter_instance.parse.assert_called_once_with(raw_content, strict=True)


@patch("dorsal.api.adapters.get_adapter")
def test_parse_file_with_file_like_object(mock_get_adapter):
    """Test parsing from a file-like object using the .read() duck-typing."""
    mock_adapter_instance = MagicMock()
    mock_adapter_instance.parse_file.return_value = {"parsed": "data"}
    mock_get_adapter.return_value = mock_adapter_instance

    fp = io.StringIO("some file content")
    result = parse_file(fp, schema_id="test/schema", source_format="txt")

    assert result == {"parsed": "data"}
    mock_adapter_instance.parse_file.assert_called_once_with(fp)


@patch("dorsal.api.adapters.get_adapter")
def test_parse_file_handles_exceptions(mock_get_adapter):
    """Test that adapter parse failures are caught and wrapped in a DorsalError."""
    mock_get_adapter.side_effect = ValueError("Corrupt file")

    with pytest.raises(DorsalError, match="Failed to parse record from fake: Corrupt file"):
        parse_file("content", schema_id="test/schema", source_format="fake")


@patch("dorsal.api.adapters.parse_file")
def test_parse_file_from_path_success(mock_parse_file, tmp_path):
    """Test standard file parsing and format inference from the input extension."""
    mock_parse_file.return_value = {"parsed": "data"}

    in_file = tmp_path / "input_test.vtt"
    in_file.write_text("dummy webvtt content", encoding="utf-8")

    result = parse_file_from_path(in_file, "test/schema")

    assert result == {"parsed": "data"}
    mock_parse_file.assert_called_once()
    assert mock_parse_file.call_args.kwargs["source_format"] == "vtt"


def test_parse_file_from_path_missing_file(tmp_path):
    """Test error handling when the requested input path does not exist."""
    missing_file = tmp_path / "ghost_file.srt"

    with pytest.raises(DorsalError, match="not a valid file or does not exist"):
        parse_file_from_path(missing_file, "test/schema")


def test_parse_file_from_path_inference_failure(tmp_path):
    """Test error handling when the format cannot be inferred from the input path."""
    in_file = tmp_path / "file_without_extension"
    in_file.write_text("dummy data", encoding="utf-8")

    with pytest.raises(DorsalError, match="Could not infer source format from file extension"):
        parse_file_from_path(in_file, "test/schema")


@patch("dorsal.api.adapters.parse_file")
def test_parse_file_from_path_dorsal_error_propagation(mock_parse_file, tmp_path):
    """Test that existing DorsalErrors raised by parse_file bubble up cleanly without double-wrapping."""
    mock_parse_file.side_effect = DorsalError("Specific internal adapter validation failed.")

    in_file = tmp_path / "input_test.txt"
    in_file.write_text("dummy data", encoding="utf-8")

    with pytest.raises(DorsalError, match="Specific internal adapter validation failed."):
        parse_file_from_path(in_file, "test/schema")


@patch("dorsal.api.adapters.parse_file")
def test_parse_file_from_path_unexpected_error(mock_parse_file, tmp_path):
    """Test that unexpected runtime exceptions are caught and wrapped cleanly."""
    mock_parse_file.side_effect = Exception("System crash")

    in_file = tmp_path / "input_test.txt"
    in_file.write_text("dummy data", encoding="utf-8")

    with pytest.raises(DorsalError, match="Failed to read or parse file"):
        parse_file_from_path(in_file, "test/schema")


@patch("dorsal.api.adapters.list_formats")
@patch.dict("dorsal.api.adapters.ALIAS_MAPPING", {"test/schema": "resolved/schema"})
def test_get_supported_formats(mock_list_formats):
    """Test that format listing proxies correctly to the registry."""
    mock_list_formats.return_value = [("srt", "SubRip Text"), ("vtt", "WebVTT")]

    result = get_supported_formats("test/schema")

    assert result == [("srt", "SubRip Text"), ("vtt", "WebVTT")]
    mock_list_formats.assert_called_once_with("resolved/schema")


@patch("dorsal.api.adapters.get_adapter")
def test_get_format_extension_success(mock_get_adapter):
    """Test that get_format_extension successfully retrieves the custom extension string."""
    mock_adapter = MagicMock()
    mock_adapter.extension = "hocr.html"
    mock_get_adapter.return_value = mock_adapter

    result = get_format_extension("test/schema", "hocr")

    assert result == "hocr.html"
    mock_get_adapter.assert_called_once_with("test/schema", "hocr")


@patch("dorsal.api.adapters.logger")
@patch("dorsal.api.adapters.get_adapter")
def test_get_format_extension_fallback(mock_get_adapter, mock_logger):
    """Test that get_format_extension safely falls back to target_format on exception."""
    mock_get_adapter.side_effect = ValueError("Adapter not found")

    result = get_format_extension("test/schema", "fake_format")

    assert result == "fake_format"

    mock_logger.warning.assert_called_once()
    assert "Adapter not found for schema" in mock_logger.warning.call_args[0][0]



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest.mock import Mock, MagicMock
import pytest
from rich.panel import Panel

from dorsal.cli.views.collection import collection_metadata


@pytest.fixture
def mock_palette():
    return {"key": "bold cyan", "primary_value": "green", "panel_title": "bold white", "panel_border": "blue"}


@pytest.fixture
def mock_collection():
    col = Mock()
    col.collection_id = "123-abc"
    col.name = "Test Collection"
    col.description = "A test description"
    col.file_count = 1000
    col.total_size = 1024 * 1024 * 50  # 50MB
    col.is_private = True
    col.date_modified = datetime.datetime(2025, 1, 1, 12, 0, 0)
    return col


def test_collection_metadata_render_full(mock_collection, mock_palette):
    """Test rendering with all fields present."""

    panel = collection_metadata(mock_collection, mock_palette)

    assert isinstance(panel, Panel)
    assert "Collection Metadata" in panel.title

    table = panel.renderable

    assert table.columns[0].style == "bold cyan"


def test_collection_metadata_no_description(mock_collection, mock_palette):
    """Test rendering when description is None."""
    mock_collection.description = None

    panel = collection_metadata(mock_collection, mock_palette)

    assert isinstance(panel, Panel)


def test_collection_metadata_no_date(mock_collection, mock_palette):
    """Test rendering when date_modified is None."""
    mock_collection.date_modified = None

    panel = collection_metadata(mock_collection, mock_palette)
    assert isinstance(panel, Panel)


def test_collection_metadata_public(mock_collection, mock_palette):
    """Test rendering for public collection."""
    mock_collection.is_private = False
    panel = collection_metadata(mock_collection, mock_palette)
    assert isinstance(panel, Panel)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, cast, List, TypeAlias, Type, TYPE_CHECKING

import jsonschema_rs

from dorsal.common.exceptions import (
    ApiDataValidationError,
    DorsalError,
    JsonSchemaValidationError,
    SchemaFormatError,
)


logger = logging.getLogger(__name__)

# See: https://json-schema.org/draft/2020-12/json-schema-validation
JSON_SCHEMA_CONSTRAINT_KEYWORDS = {
    "$ref",
    "allOf",
    "anyOf",
    "oneOf",
    "not",
    "if",
    "then",
    "else",
    "dependentSchemas",
    "properties",
    "patternProperties",
    "additionalProperties",
    "items",
    "additionalItems",
    "contains",
    "propertyNames",
    "type",
    "enum",
    "const",
    "multipleOf",
    "maximum",
    "exclusiveMaximum",
    "minimum",
    "exclusiveMinimum",
    "maxLength",
    "minLength",
    "pattern",
    "maxItems",
    "minItems",
    "uniqueItems",
    "maxContains",
    "minContains",
    "maxProperties",
    "minProperties",
    "required",
    "dependentRequired",
    "format",
    "contentEncoding",
    "contentMediaType",
    "contentSchema",
}

if TYPE_CHECKING:
    from jsonschema_rs import Validator as JsonSchemaRustValidator


class JsonSchemaValidator:
    """
    A wrapper around the jsonschema-rs validator.

    This class serves as a Carrier for the compiled Rust validator engine,
    while retaining access to the original schema dictionary (required by
    ModelRunner) and providing a concrete type for static analysis.
    """

    __slots__ = ("schema", "validator", "__name__")

    def __init__(self, schema: dict, strict: bool = True):
        self._validate_schema_structure(schema, strict)

        self.schema = schema

        title = schema.get("title", "Untitled")
        self.__name__ = f"JsonSchemaValidator[{title}]"

        try:
            self.validator: JsonSchemaRustValidator = jsonschema_rs.validator_for(schema, validate_formats=True)
        except Exception as err:
            logger.exception(
                "The provided schema is structurally invalid and cannot be used to create a validator: %s",
                err,
            )
            raise SchemaFormatError(
                message="The provided schema is invalid and cannot be used to prepare a validator.",
                schema_error_detail=str(err),
            ) from err

    def _validate_schema_structure(self, schema: Any, strict: bool) -> None:
        """Internal helper to validate input schema before compilation."""
        if not isinstance(schema, dict):
            logger.error("Schema must be a dictionary. Got type: %s", type(schema).__name__)
            raise TypeError(f"The 'schema' argument must be a dictionary, got {type(schema).__name__}.")

        if not schema:
            logger.error("Schema dictionary cannot be empty.")
            raise ValueError("The 'schema' dictionary cannot be empty.")

        if strict:
            if not any(key in schema for key in JSON_SCHEMA_CONSTRAINT_KEYWORDS):
                logger.warning("Schema appears to be inert (no constraint keywords found).")
                raise ValueError(
                    "The provided schema is inert: it contains no known JSON Schema "
                    "constraint keywords (like 'type', 'properties', '$ref', etc.) "
                    "and would silently pass all validation."
                )

    def validate(self, instance: Any) -> None:
        """
        Validate the instance, raising jsonschema_rs.ValidationError on failure.
        """
        self.validator.validate(instance)

    def is_valid(self, instance: Any) -> bool:
        """Return True if valid, False otherwise."""
        return self.validator.is_valid(instance)

    def __call__(self, instance: Any) -> None:
        return self.validate(instance)


def get_json_schema_validator(schema: dict, strict: bool = True) -> JsonSchemaValidator:
    """Create a JsonSchemaValidator instance."""
    return JsonSchemaValidator(schema, strict=strict)


def json_schema_validate_records(records: list[dict] | Any, validator: JsonSchemaValidator) -> dict:
    """Validates records using a pre-configured jsonschema-rs validator."""
    validator_name = getattr(validator, "__name__", "JsonSchemaValidator")

    logger.debug(
        "Validating %s records with Rust validator '%s'.",
        len(records) if isinstance(records, list) else "an unknown number of",
        validator_name,
    )

    if not isinstance(records, list):
        logger.warning("Input 'records' must be a list. Got: %s", type(records).__name__)
        raise ValueError(f"Input 'records' must be a list, got {type(records).__name__}.")

    if not records:
        return {
            "total_records": 0,
            "valid_records": 0,
            "invalid_records": 0,
            "error_details": [],
        }

    valid_records_count = 0
    error_details_list = []

    for index, record in enumerate(records):
        if not isinstance(record, dict):
            error_details_list.append(
                {
                    "record_index": index,
                    "record_preview": str(record)[:100],
                    "error_message": f"Invalid record type: {type(record).__name__}",
                    "path": [],
                    "validator": "type_check",
                }
            )
            continue

        try:
            validator.validate(record)
            valid_records_count += 1
        except JsonSchemaValidationError as err:
            record_str = str(record)
            record_preview = record_str[:150] + "..." if len(record_str) > 150 else record_str

            path_list = [str(p) for p in err.instance_path]

            error_details_list.append(
                {
                    "record_index": index,
                    "record_preview": record_preview,
                    "error_message": err.message,
                    "path": path_list,
                    "validator": validator_name,
                }
            )
            logger.debug(
                "Record at index %d failed validation against %s: %s (Path: %s)",
                index,
                validator_name,
                err.message,
                path_list,
            )
        except Exception as err:
            record_str = str(record)
            record_preview = record_str[:150] + "..." if len(record_str) > 150 else record_str
            error_details_list.append(
                {
                    "record_index": index,
                    "record_preview": record_preview,
                    "error_message": f"An unexpected error occurred during this record's validation: {str(err)}",
                    "path": [],
                    "validator": "unknown_error",
                }
            )
            logger.exception("Unexpected error validating record at index %d.", index)

    invalid_records_count = len(records) - valid_records_count

    summary = {
        "total_records": len(records),
        "valid_records": valid_records_count,
        "invalid_records": invalid_records_count,
        "error_details": error_details_list,
    }

    if invalid_records_count > 0:
        logger.warning(
            "Validation completed: %d valid, %d invalid out of %d total records.",
            valid_records_count,
            invalid_records_count,
            len(records),
        )
    else:
        logger.debug("Validation completed: All %d records are valid.", len(records))

    return summary



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated, Optional
import typer
import pathlib

from dorsal.cli.search_app import search_and_display


def search_record(
    ctx: typer.Context,
    query: Annotated[
        str,
        typer.Argument(help="The search query string. Queries with spaces must be enclosed in quotes."),
    ],
    is_global: Annotated[
        bool,
        typer.Option(
            "--global",
            "-g",
            help="Perform a global search across all public files (Premium).",
            rich_help_panel="Search Options",
        ),
    ] = False,
    page: Annotated[
        int,
        typer.Option(
            "--page",
            "-p",
            help="The page number of results to display.",
            rich_help_panel="Search Options",
        ),
    ] = 1,
    per_page: Annotated[
        int,
        typer.Option(
            "--per-page",
            help="The number of results to display per page.",
            rich_help_panel="Search Options",
        ),
    ] = 30,
    sort_by: Annotated[
        str,
        typer.Option(
            "--sort-by",
            help="Field to sort results by.",
            rich_help_panel="Search Options",
        ),
    ] = "date_modified",
    sort_order: Annotated[
        str,
        typer.Option(
            "--sort-order",
            help="Sort order ('asc' or 'desc').",
            rich_help_panel="Search Options",
        ),
    ] = "desc",
    match_any: Annotated[
        bool,
        typer.Option(
            "--or",
            help="Use OR logic for the query. By default, multiple terms are combined with AND.",
            rich_help_panel="Search Options",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output results as a raw JSON object.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    save: Annotated[
        bool,
        typer.Option(
            "-s",
            "--save",
            help="Save the JSON search results to the default directory or --output path.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    output_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "-o",
            "--output",
            help="Custom path to save the JSON search results (e.g., 'results.json').",
            dir_okay=True,
            file_okay=True,
            writable=True,
            resolve_path=True,
            rich_help_panel="Output Options",
        ),
    ] = None,
):
    """
    Search DorsalHub file records.
    """
    scope = "global" if is_global else "user"

    search_and_display(
        ctx=ctx,
        scope=scope,
        query=query,
        page=page,
        per_page=per_page,
        sort_by=sort_by,
        sort_order=sort_order,
        json_output=json_output,
        match_any=match_any,
        save=save,
        output_path=output_path,
    )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer

from dorsal.cli.record_app.get_cmd import get_file_record
from dorsal.cli.record_app.delete_cmd import delete_file_record
from dorsal.cli.record_app.tag_app import tag_app
from dorsal.cli.record_app.search_cmd import search_record

app = typer.Typer(
    name="record",
    help="Search, view and delete file metadata records and tags on DorsalHub.",
    no_args_is_help=True,
)

app.command(name="get")(get_file_record)
app.command(name="delete")(delete_file_record)
app.command(name="search")(search_record)
app.add_typer(tag_app, name="tag")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import typer
from rich.console import Group, RenderableType
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from typing import Dict, Annotated
import json

app = typer.Typer(
    name="config",
    help="View or manage Dorsal settings.",
    no_args_is_help=True,
)

theme_app = typer.Typer(name="theme", help="List and set color themes.", no_args_is_help=True)
app.add_typer(theme_app)

pipeline_app = typer.Typer(
    name="pipeline", help="Inspect and manage the annotation model pipeline.", no_args_is_help=True
)
app.add_typer(pipeline_app)


@app.command(name="show")
def show_config(
    ctx: typer.Context,
    json_output: Annotated[
        bool,
        typer.Option("--json", help="Output the configuration as a raw JSON object."),
    ] = False,
):
    """
    Displays the current configuration status and paths.
    """
    from dorsal.common.cli import get_rich_console
    from dorsal.api.config import get_config_summary

    console = get_rich_console()
    palette = ctx.obj["palette"]

    config_data = get_config_summary()

    if json_output:
        console.print(json.dumps(config_data, indent=2, default=str))
        raise typer.Exit()

    if config_data["api_key_set"]:
        status_text = Text(f"Set (from {config_data['api_key_source']})", style=palette.get("success", "green"))
    else:
        status_text = Text("Not Set", style=palette.get("warning", "yellow"))

    user_text = (
        Text(config_data["logged_in_user"], style=palette.get("primary_value", "default"))
        if config_data["logged_in_user"]
        else Text("N/A (run 'dorsal auth login')", style=palette.get("info", "dim"))
    )

    config_table = Table.grid(expand=True, padding=(0, 2))
    config_table.add_column(justify="right", style=palette.get("key", "dim"), width=22)
    config_table.add_column()

    config_table.add_row(
        "Current Theme:",
        Text(config_data["current_theme"], style=palette.get("primary_value", "default")),
    )
    config_table.add_row("Logged-In User:", user_text)
    config_table.add_row("API Key Status:", status_text)
    config_table.add_row(
        "API URL:",
        Text(config_data["api_url"], style=palette.get("primary_value", "default")),
    )
    config_table.add_row(
        "Reports Path:",
        Text(config_data["reports_path"], style=palette.get("primary_value", "default")),
    )
    config_table.add_row(
        "Active Config File:",
        Text(config_data["active_config_path"], style=palette.get("primary_value", "default")),
    )
    config_table.add_row(
        "Global Config File:",
        Text(config_data["global_config_path"], style=palette.get("primary_value", "default")),
    )

    console.print(
        Panel(
            config_table,
            title=f"[{palette.get('panel_title', 'bold')}]Dorsal Configuration[/]",
            border_style=palette.get("panel_border", "default"),
            expand=False,
        )
    )

    console.print("\n[dim]For more detail on the pipeline config, run:[/]")
    console.print("[dim]  dorsal config pipeline show[/]")


def _create_theme_preview_panel(theme_name: str, palette: dict[str, str]) -> Panel:
    border = palette.get("panel_border_alt", palette.get("panel_border", "default"))
    title_style = palette.get("panel_title_alt", palette.get("panel_title", "default"))
    preview_grid = Table.grid(padding=(0, 2))
    preview_grid.add_column(style=palette.get("key", "dim"), width=12)
    preview_grid.add_column()
    preview_grid.add_row("Key:", Text("Primary Value", style=palette.get("primary_value", "default")))
    preview_grid.add_row("Hash:", Text("a1b2c3d4e5f62a3b4c...", style=palette.get("hash_value", "default")))
    public_tag_text = Text.assemble(
        ("public_tag", palette.get("tag_public", "default")),
        (" (public)", palette.get("tag_subtext", "default")),
    )
    private_tag_text = Text.assemble(
        ("private_tag", palette.get("tag_private", "default")),
        (" (private)", palette.get("tag_subtext", "default")),
    )
    preview_grid.add_row("Tags:", public_tag_text)
    preview_grid.add_row("", private_tag_text)
    return Panel(
        preview_grid,
        title=f"[{title_style}]{theme_name}[/]",
        border_style=border,
        expand=False,
    )


@theme_app.command(name="list")
def list_themes(ctx: typer.Context):
    """Lists all available built-in and custom color themes."""
    from dorsal.common.cli import get_rich_console
    from dorsal.cli.themes.palettes import BUILT_IN_PALETTES, _load_custom_palettes

    console = get_rich_console()

    palette = ctx.obj["palette"]
    renderables: list[RenderableType] = [Text.from_markup(f"[{palette.get('panel_title')}]🎨 Available Themes[/]")]

    renderables.append(Text.from_markup(f"\n[{palette.get('section_title')}]Built-in Themes[/]"))
    for name, theme_palette in BUILT_IN_PALETTES.items():
        renderables.append(_create_theme_preview_panel(name, theme_palette))

    custom = _load_custom_palettes()
    if custom:
        renderables.append(Text(""))
        renderables.append(
            Text.from_markup(f"[{palette.get('section_title')}]Custom Themes[/] (from ~/.dorsal/palettes.json)")
        )
        for name, theme_palette in custom.items():
            renderables.append(_create_theme_preview_panel(name, theme_palette))

    command_color = palette.get("primary_value", "default")
    renderables.append(
        Text.from_markup(
            f"\n[bold {command_color}]dorsal config theme set <name>[/] to set a theme.\n\n"
            f"Or use the --theme flag before another command\ne.g. [bold {command_color}]dorsal --theme mono file push ...[/]`"
        )
    )
    console.print(
        Panel(
            Group(*renderables),
            expand=False,
            title=f"[{palette.get('panel_title')}]Theme Configuration[/]",
            border_style=palette.get("panel_border", "blue"),
            padding=(1, 2),
        )
    )


@theme_app.command(name="set")
def set_theme(
    ctx: typer.Context,
    theme_name: str = typer.Argument(..., help="The name of the theme to set as the default."),
):
    """Sets the default color theme, saved to the config file."""
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR
    from dorsal.common.auth import write_theme_to_config
    from dorsal.cli.themes.palettes import BUILT_IN_PALETTES, _load_custom_palettes

    console = get_rich_console()

    palette = ctx.obj["palette"]
    all_palettes = {**BUILT_IN_PALETTES, **_load_custom_palettes()}
    if theme_name not in all_palettes:
        error_message = f"[{palette.get('error', 'red')}]Error:[/] Theme '{theme_name}' not found. Use `dorsal config theme list` to see available themes."
        console.print(error_message)
        exit_cli(code=EXIT_CODE_ERROR)

    try:
        write_theme_to_config(theme_name)
        success_message = Text.assemble(
            ("✅ Default theme set to '", palette.get("success", "default")),
            (theme_name, f"bold {palette.get('primary_value', 'default')}"),
            ("'.", palette.get("success", "green")),
        )
        console.print(success_message)
    except OSError as e:
        exit_cli(code=EXIT_CODE_ERROR, message=f"Error saving configuration: {e}")


def _handle_pipeline_action(target: str, func_index, func_name, action_desc: str):
    """
    Helper to dispatch commands to either index-based or name-based API functions.
    """
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR

    console = get_rich_console()

    try:
        idx = int(target)
        func_index(index=idx)
        console.print(f"[green]Successfully {action_desc} model at index {idx}.[/]")
    except ValueError:
        try:
            func_name(name=target)
            console.print(f"[green]Successfully {action_desc} model '{target}'.[/]")
        except Exception as e:
            console.print(f"[red]Error: {e} [/]")
            exit_cli(EXIT_CODE_ERROR)
    except Exception as e:
        console.print(f"[red]Error: {e} [/]")
        exit_cli(EXIT_CODE_ERROR)


@pipeline_app.command(name="show")
def show_pipeline(ctx: typer.Context):
    """
    Show the current effective pipeline configuration.
    """
    from dorsal.common.cli import get_rich_console
    from dorsal.api.config import show_model_pipeline

    console = get_rich_console()
    palette = ctx.obj["palette"]

    summary = show_model_pipeline()

    if not summary:
        console.print(Panel("The pipeline is currently empty.", title="Pipeline", border_style=palette.get("warning")))
        return

    table = Table(expand=True, box=None, padding=(0, 2))
    table.add_column("Idx", justify="right", style="dim")
    table.add_column("Status", justify="left", width=10)
    table.add_column("Model Name", style="bold")
    table.add_column("Module", style="dim")
    table.add_column("Schema ID", style="cyan")
    table.add_column("Dependencies", style="dim")

    for step in summary:
        status = "✅ Active"
        name = step["name"]

        if step["status"] == "Deactivated":
            status = "❌ Deactivated"
        elif step["status"] == "Base (Locked)":
            status = "[dim]🔒 Default[/]"
            name = f"[dim]{step['name']}[/]"

        table.add_row(str(step["index"]), status, name, step["module"], step["schema_id"], step["dependencies"])

    console.print(
        Panel(
            table,
            expand=False,
            title=f"[{palette.get('panel_title')}]Annotation Model Pipeline[/]",
            border_style=palette.get("panel_border", "blue"),
            subtitle=f"Total Models: {len(summary)}",
        )
    )


@pipeline_app.command(name="remove")
def remove_step(target: str = typer.Argument(..., help="The index or name of the model to remove.")):
    """Remove a model from the pipeline.

    Note: there is currently no way to remove entries from the global config via the CLI. Use the Python API for that.

    """
    from dorsal.api.config import remove_model_by_index, remove_model_by_name

    _handle_pipeline_action(target, remove_model_by_index, remove_model_by_name, "removed")


@pipeline_app.command(name="activate")
def activate_step(target: str = typer.Argument(..., help="The index or name of the model to activate.")):
    """Enable a previously deactivated model."""
    from dorsal.api.config import activate_model_by_index, activate_model_by_name

    _handle_pipeline_action(target, activate_model_by_index, activate_model_by_name, "activated")


@pipeline_app.command(name="deactivate")
def deactivate_step(target: str = typer.Argument(..., help="The index or name of the model to deactivate.")):
    """Disable a model without removing it from the config."""
    from dorsal.api.config import deactivate_model_by_index, deactivate_model_by_name

    _handle_pipeline_action(target, deactivate_model_by_index, deactivate_model_by_name, "deactivated")


@pipeline_app.command(name="check")
def check_pipeline(
    ctx: typer.Context,
    fix: bool = typer.Option(False, "--fix", help="Automatically remove models with broken import paths."),
):
    """
    Verifies that all registered models can be imported.
    """
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR
    from dorsal.common.validators import import_callable, CallableImportPath
    from dorsal.api.config import get_model_pipeline, remove_model_by_index

    console = get_rich_console()
    palette = ctx.obj["palette"]

    steps = get_model_pipeline(scope="effective")
    broken_indices = []

    with console.status("Checking pipeline integrity..."):
        for i, step in enumerate(steps):
            if i == 0:
                continue

            model_path = step.annotation_model
            try:
                import_callable(model_path)
                if isinstance(step.validation_model, CallableImportPath):
                    import_callable(step.validation_model)
            except (ImportError, ModuleNotFoundError, AttributeError) as e:
                broken_indices.append((i, step.annotation_model.name, str(e)))

    if not broken_indices:
        console.print(f"[{palette.get('success', 'green')}]✓ All pipeline models are importable.[/]")
        raise typer.Exit()

    console.print(f"[{palette.get('error', 'red')}]Found {len(broken_indices)} broken models:[/]")
    for idx, name, err in broken_indices:
        console.print(f"  • Index {idx}: [bold]{name}[/] - {err}")

    if fix:
        for idx, name, _ in reversed(broken_indices):
            try:
                remove_model_by_index(index=idx)
                console.print(f"  [yellow]Removed broken model '{name}' (Index {idx})[/]")
            except Exception as e:
                console.print(f"  [red]Failed to remove index {idx}: {e}[/]")

        console.print(f"\n[{palette.get('success', 'green')}]Cleanup complete.[/]")
    else:
        console.print("\n[dim]Run with [white]--fix[/] to automatically remove these models.[/]")
        exit_cli(EXIT_CODE_ERROR)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import zipfile
import xml.etree.ElementTree as ET
from typing import Any, List, Dict, Optional
from datetime import datetime

from dorsal.common.language import normalize_language_name, extract_locale_code, normalize_language_alpha3
from dorsal.file.utils.dates import PDF_DATETIME

logger = logging.getLogger(__name__)

NS_MAP = {
    "c": "urn:oasis:names:tc:opendocument:xmlns:container",
    "opf": "http://www.idpf.org/2007/opf",
    "dc": "http://purl.org/dc/elements/1.1/",
}

HTML_TAG_RX = re.compile(r"<[^>]+>")
ISBN_RX = re.compile(r"(?:ISBN[ \t]*:?|urn:isbn:)\s*([0-9]{9,12}[0-9X])")
CLEAN_ISBN_RX = re.compile(r"^([0-9]{9}[0-9X]|[0-9]{13})$")
PLACEHOLDER_DATE_RX = re.compile(r"^(0001|0101)-01-01")
TOOL_RX = re.compile(
    r"^(calibre|ibooks author|smashwords|adobe|kindle|pressbooks|libreoffice|easypress|innodata|pyscript)",
    re.IGNORECASE,
)


def _strip_html(text: str | None) -> str | None:
    """Removes HTML tags from a string. Returns None if result is empty."""
    if not text:
        return None
    cleaned_text = HTML_TAG_RX.sub("", text).strip()
    return cleaned_text if cleaned_text else None


def _parse_date(date_str: str | None) -> datetime | None:
    """Parses a date string, using the PDFDatetime parser."""
    if not date_str or not PDF_DATETIME:
        return None

    return PDF_DATETIME.parse(date_str)


def _extract_isbn(id_str: str | None) -> str | None:
    """Extracts a valid ISBN-10 or ISBN-13, stripping prefixes."""
    if not id_str:
        return None

    match = ISBN_RX.search(id_str)
    if match:
        return match.group(1).replace("-", "")

    clean_id = id_str.replace("-", "").strip()
    if CLEAN_ISBN_RX.match(clean_id):
        return clean_id

    return None


def _get_meta_text(metadata_el: ET.Element, tag: str) -> str | None:
    """Find a single metadata tag's text (namespace-agnostic)."""
    el = metadata_el.find(f"dc:{tag}", NS_MAP)
    if el is not None and el.text:
        return el.text.strip()
    el = metadata_el.find(tag)
    return el.text.strip() if el is not None and el.text else None


def _get_meta_list(metadata_el: ET.Element, tag: str) -> List[str]:
    """Find all metadata tags (namespace-agnostic) and return text list."""
    elements = metadata_el.findall(f"dc:{tag}", NS_MAP)
    if not elements:
        elements = metadata_el.findall(tag)
    if not elements:
        return []
    return [el.text for el in elements if el.text]


def _get_all_epub_dates(metadata_el: ET.Element) -> dict[str, str | None]:
    """
    Parses all dc:date elements and sorts them by their opf:event.
    """
    event_key = f"{{{NS_MAP['opf']}}}event"
    date_els = metadata_el.findall("dc:date", NS_MAP)
    if not date_els:
        date_els = metadata_el.findall("date")

    dates: dict[str, str | None] = {
        "publication": None,
        "creation": None,
        "modification": None,
        "first_unspecified": None,
    }

    for el in date_els:
        date_str = el.text.strip() if el.text else None
        if not date_str:
            continue

        event = el.get(event_key)

        if event == "publication" and not dates["publication"]:
            dates["publication"] = date_str
        elif event == "modification" and not dates["modification"]:
            dates["modification"] = date_str
        elif event == "creation" and not dates["creation"]:
            dates["creation"] = date_str
        elif event is None and not dates["first_unspecified"]:
            dates["first_unspecified"] = date_str

    return dates


def _get_cover_path(opf_root: ET.Element) -> str | None:
    """
    Finds the cover image path from the <metadata> and <manifest>.
    """
    metadata_el = opf_root.find("opf:metadata", NS_MAP)
    if metadata_el is None:
        return None

    cover_id = None
    for meta_el in metadata_el.findall("opf:meta", NS_MAP) + metadata_el.findall("meta"):
        if meta_el.get("name") == "cover":
            cover_id = meta_el.get("content")
            break

    if not cover_id:
        logger.debug('Could not find <meta name="cover"> tag in metadata.')
        return None

    manifest_el = opf_root.find("opf:manifest", NS_MAP)
    if manifest_el is None:
        logger.debug("Could not find <manifest> element in .opf.")
        return None

    search_paths = [f"opf:item[@id='{cover_id}']", f"item[@id='{cover_id}']"]
    cover_item = None
    for path in search_paths:
        cover_item = manifest_el.find(path, NS_MAP)
        if cover_item is not None:
            break

    if cover_item is None:
        logger.debug("Found cover ID '%s' but no matching <item> in manifest.", cover_id)
        return None

    cover_href = cover_item.get("href")
    if not cover_href:
        logger.debug("Cover <item> '%s' has no 'href' attribute.", cover_id)
        return None

    return cover_href


def extract_epub_metadata(file_path: str) -> dict[str, Any] | None:
    """
    Extracts and normalizes metadata from an EPUB file.
    """
    try:
        with zipfile.ZipFile(file_path, "r") as z:
            try:
                container_data = z.read("META-INF/container.xml")
            except KeyError:
                logger.debug("META-INF/container.xml not found in '%s'.", file_path)
                return None
            container_root = ET.fromstring(container_data)
            opf_path_el = container_root.find("c:rootfiles/c:rootfile", NS_MAP)
            if opf_path_el is None:
                logger.debug("Could not find .opf file path in container.xml for '%s'.", file_path)
                return None
            opf_path = opf_path_el.get("full-path")
            if not opf_path:
                logger.debug("Invalid container.xml in '%s' (no full-path).", file_path)
                return None

            opf_data = z.read(opf_path)
            opf_root = ET.fromstring(opf_data)

            metadata_el = opf_root.find("opf:metadata", NS_MAP)
            if metadata_el is None:
                logger.debug("Could not find <metadata> element in .opf for '%s'.", file_path)
                return None

            raw_identifiers = _get_meta_list(metadata_el, "identifier")
            isbn = None
            other_identifiers = []

            for id_str in raw_identifiers:
                if not id_str:
                    continue

                found_isbn = _extract_isbn(id_str)
                if found_isbn and not isbn:
                    isbn = found_isbn
                else:
                    other_identifiers.append(id_str.strip())

            tool_name_from_meta = None
            for meta_el in metadata_el.findall("opf:meta", NS_MAP) + metadata_el.findall("meta"):
                if meta_el.get("name") in ("generator", "producer"):
                    tool_name_from_meta = meta_el.get("content")
                    if tool_name_from_meta:
                        break

            raw_contributors = _get_meta_list(metadata_el, "contributor")
            final_contributors = []
            final_tools = []

            if tool_name_from_meta:
                final_tools.append(tool_name_from_meta.strip())

            for entry in raw_contributors:
                if not entry:
                    continue

                if TOOL_RX.search(entry):
                    final_tools.append(entry.strip())
                else:
                    final_contributors.append(entry.strip())

            all_dates = _get_all_epub_dates(metadata_el)
            pub_date_str = (
                all_dates.get("publication")
                or all_dates.get("creation")
                or all_dates.get("first_unspecified")
                or all_dates.get("modification")
            )

            raw_lang_str = _get_meta_text(metadata_el, "language")

            metadata = {
                "title": _get_meta_text(metadata_el, "title"),
                "authors": _get_meta_list(metadata_el, "creator"),
                "contributors": final_contributors,
                "publisher": _get_meta_text(metadata_el, "publisher"),
                "language": normalize_language_name(raw_lang_str),
                "language_code": normalize_language_alpha3(raw_lang_str),
                "locale_code": extract_locale_code(raw_lang_str),
                "subjects": _get_meta_list(metadata_el, "subject"),
                "description": _strip_html(_get_meta_text(metadata_el, "description")),
                "rights": _get_meta_text(metadata_el, "rights"),
                "isbn": isbn,
                "other_identifiers": other_identifiers,
                "cover_path": _get_cover_path(opf_root),
                "tools": list(set(final_tools)),
                "publication_date": _parse_date(pub_date_str),
                "creation_date": _parse_date(all_dates.get("creation")),
                "modification_date": _parse_date(all_dates.get("modification")),
            }

            return metadata

    except zipfile.BadZipFile:
        logger.warning("File is not a valid ZIP archive (may be corrupt or DRM-protected): '%s'", file_path)
        return None
    except ET.ParseError:
        logger.warning("Failed to parse XML from ebook: '%s'", file_path)
        return None
    except Exception as e:
        logger.warning("Failed to parse EPUB file '%s': %s", file_path, e, exc_info=True)
        return None


def extract_mobi_metadata(file_path: str) -> dict[str, Any] | None:
    """STUB for extracting metadata from a MOBI file."""
    logger.warning("MOBI metadata extraction is not implemented (file: %s).", file_path)
    return None



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import typer
from typing import Annotated
import json

from rich.panel import Panel
from rich.table import Table

logger = logging.getLogger(__name__)


def optimize_cache(
    ctx: typer.Context,
    json_output: Annotated[
        bool,
        typer.Option("--json", help="Output the optimization summary as a raw JSON object."),
    ] = False,
):
    """
    Runs a full maintenance routine on the cache.

    This includes:
    - Pruning stale records (for deleted or modified files).
    - Syncing record compression with the current config.
    - Reclaiming disk space (VACUUM).
    """
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR
    from dorsal.session import get_shared_cache
    from dorsal.file.utils.size import human_filesize

    console = get_rich_console()
    palette: dict[str, str] = ctx.obj["palette"]

    try:
        with console.status("Optimizing cache... (this may take a moment)"):
            cache = get_shared_cache()
            results = cache.optimize()

        if json_output:
            console.print(json.dumps(results))
            exit_cli()

        summary_table = Table.grid(expand=False)
        summary_table.add_column(justify="right", style=palette.get("key", "dim"), width=35)
        summary_table.add_column(justify="left", style=palette.get("primary_value", "default"))

        summary_table.add_row("Records Removed:", f"{results.get('stale_records_removed', 0):,}")
        summary_table.add_row(
            "Records Reprocessed:",
            f"{results.get('records_rewritten_for_compression', 0):,}",
        )
        summary_table.add_row()
        summary_table.add_row(
            "Size Before Optimization:",
            human_filesize(results.get("size_before_bytes", 0)),
        )
        summary_table.add_row(
            "Size After Optimization:",
            human_filesize(results.get("size_after_bytes", 0)),
        )
        summary_table.add_row(
            "Disk Space Reclaimed:",
            human_filesize(results.get("size_reclaimed_bytes", 0)),
        )

        console.print(
            Panel(
                summary_table,
                title=f"[{palette.get('panel_title_success', 'bold green')}]✅ Cache Optimization Complete[/]",
                border_style=palette.get("panel_border_success", "green"),
                expand=False,
            )
        )

    except typer.Exit:
        raise
    except Exception as err:
        logger.exception("Failed to optimize cache.")
        exit_cli(
            code=EXIT_CODE_ERROR,
            message=f"An error occurred while optimizing the cache: {err}",
        )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from unittest.mock import Mock, MagicMock, patch, call
import logging
import sys

from dorsal.file.preprocessing.pdf import (
    extract_pdf_layout_per_mille,
    extract_pdf_layout_normalized,
    extract_pdf_layout_pts,
    extract_pdf_pages,
    ocr_extract_pdf_text,
    DependencyError,
    PDFProcessingError,
    PDFPage,
    PDFToken,
)


@pytest.fixture
def mock_dependencies():
    """
    Patches 'pypdfium2' and 'PIL' in sys.modules.
    Returns (mock_pypdfium, mock_pillow).
    """
    mock_pdfium = MagicMock()
    mock_pdfium.PdfiumError = type("PdfiumError", (Exception,), {})
    mock_pil = MagicMock()

    with patch.dict(sys.modules, {"pypdfium2": mock_pdfium, "PIL": mock_pil}):
        yield mock_pdfium, mock_pil


@pytest.fixture
def dummy_pdf_path():
    return "dummy.pdf"


def create_mock_page(width=100, height=200, rects=None):
    """Helper to create a mock PDF page with specific geometry."""
    mock_page = Mock()
    mock_textpage = Mock()

    mock_page.get_size.return_value = (width, height)
    mock_page.get_textpage.return_value = mock_textpage

    if rects:
        mock_textpage.count_rects.return_value = len(rects)
        # side_effect allows iterating through the rects list on subsequent calls
        mock_textpage.get_rect.side_effect = [r["rect"] for r in rects]
        mock_textpage.get_text_bounded.side_effect = [r["text"] for r in rects]
    else:
        mock_textpage.count_rects.return_value = 0

    mock_textpage.get_text_range.return_value = " ".join([r["text"] for r in rects]) if rects else ""
    return mock_page


# --- Tests for extract_pdf_layout_per_mille ---


class TestExtractLayoutPerMille:
    def test_per_mille_math_and_types(self, mock_dependencies, dummy_pdf_path):
        """Test scaling to 1000 and integer casting."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()

        # Setup Page: 100x200
        # Rect: Left=10, Bottom=180, Right=50, Top=190
        # PDF Origin is Bottom-Left.
        # Top-Left conversion:
        #   y0 = 200 - 190 = 10
        #   y1 = 200 - 180 = 20
        # Scale (1000):
        #   x0 = (10/100)*1000 = 100
        #   y0 = (10/200)*1000 = 50
        #   x1 = (50/100)*1000 = 500
        #   y1 = (20/200)*1000 = 100
        rect_data = [{"rect": (10, 180, 50, 190), "text": "Header"}]
        mock_page = create_mock_page(100, 200, rect_data)

        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__iter__ = Mock(return_value=iter([mock_page]))

        results = extract_pdf_layout_per_mille(dummy_pdf_path)

        assert len(results) == 1
        page = results[0]

        # Check Types
        assert isinstance(page, PDFPage)
        assert len(page.tokens) == 1
        token = page.tokens[0]
        assert isinstance(token, PDFToken)

        # Check Values
        assert token.text == "Header"
        assert token.box == (100, 50, 500, 100)

        # Verify strict Int typing
        for coord in token.box:
            assert isinstance(coord, int)

    def test_strict_mode_raises_on_invalid_dimensions(self, mock_dependencies, dummy_pdf_path):
        """Test strict=True raises error on 0x0 page."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()

        # Page with 0 width/height
        mock_page = create_mock_page(0, 0, [])
        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__iter__ = Mock(return_value=iter([mock_page]))

        with pytest.raises(PDFProcessingError) as exc:
            extract_pdf_layout_per_mille(dummy_pdf_path, strict=True)

        assert "invalid dimensions" in str(exc.value)

    def test_lax_mode_clamps_invalid_dimensions(self, mock_dependencies, dummy_pdf_path):
        """Test strict=False (default) clamps values safely."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()

        # Page 0x0. Logic uses 'else 1' for divisor.
        # Rect: 10, 10, 20, 20 (Left, Bottom, Right, Top)
        # X calculation: 10/1 * 1000 = 10000 -> Clamped to 1000
        # Y calculation (Flipped): 0 (height) - 20 (top) = -20 -> Clamped to 0 by max(0, ...)
        rect_data = [{"rect": (10, 10, 20, 20), "text": "Foo"}]
        mock_page = create_mock_page(0, 0, rect_data)

        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__iter__ = Mock(return_value=iter([mock_page]))

        results = extract_pdf_layout_per_mille(dummy_pdf_path, strict=False)

        assert len(results) == 1
        # X is clamped to max (1000), Y is clamped to min (0) due to negative flip
        assert results[0].tokens[0].box == (1000, 0, 1000, 0)

    def test_input_floats_forced_to_ints(self, mock_dependencies, dummy_pdf_path):
        """Verify that if pdfium returns floats, we strictly cast to int."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()

        # Rect returns floats
        rect_data = [{"rect": (10.5, 180.5, 50.5, 190.5), "text": "Floaty"}]
        mock_page = create_mock_page(100, 200, rect_data)

        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__iter__ = Mock(return_value=iter([mock_page]))

        results = extract_pdf_layout_per_mille(dummy_pdf_path)
        token = results[0].tokens[0]

        # Box values should be integers, not floats
        for val in token.box:
            assert isinstance(val, int)


# --- Tests for extract_pdf_layout_normalized ---


class TestExtractLayoutNormalized:
    def test_normalized_math(self, mock_dependencies, dummy_pdf_path):
        """Test scaling to 0.0-1.0 floats."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()

        # Page: 100x200
        # Rect: L=10, B=180, R=50, T=190
        # Y-Flip: T=10, B=20
        # Normalized:
        # x0 = 10/100 = 0.1
        # y0 = 10/200 = 0.05
        # x1 = 50/100 = 0.5
        # y1 = 20/200 = 0.1
        rect_data = [{"rect": (10, 180, 50, 190), "text": "Norm"}]
        mock_page = create_mock_page(100, 200, rect_data)

        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__iter__ = Mock(return_value=iter([mock_page]))

        results = extract_pdf_layout_normalized(dummy_pdf_path)
        token = results[0].tokens[0]

        assert token.box == (0.1, 0.05, 0.5, 0.1)
        for val in token.box:
            assert isinstance(val, float)

    def test_strict_mode_normalized(self, mock_dependencies, dummy_pdf_path):
        """Test strict mode prevents garbage normalization on 0x0 pages."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()
        mock_page = create_mock_page(0, 0, [])
        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__iter__ = Mock(return_value=iter([mock_page]))

        with pytest.raises(PDFProcessingError):
            extract_pdf_layout_normalized(dummy_pdf_path, strict=True)


# --- Tests for extract_pdf_layout_pts ---


class TestExtractLayoutPts:
    def test_pts_math_no_scaling(self, mock_dependencies, dummy_pdf_path):
        """Test raw PTs (only Y-flip, no scaling)."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()

        # Page: 100x200
        # Rect: L=10, B=180, R=50, T=190
        # Expected (Top-Left Origin):
        # x0 = 10
        # y0 = 200 - 190 = 10
        # x1 = 50
        # y1 = 200 - 180 = 20
        rect_data = [{"rect": (10, 180, 50, 190), "text": "Points"}]
        mock_page = create_mock_page(100, 200, rect_data)

        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__iter__ = Mock(return_value=iter([mock_page]))

        results = extract_pdf_layout_pts(dummy_pdf_path)
        token = results[0].tokens[0]

        # Should be floats, matched exactly to coordinates
        assert token.box == (10.0, 10.0, 50.0, 20.0)


# --- Common Exception Handling ---


class TestCommonErrors:
    def test_missing_dependency(self, dummy_pdf_path):
        with patch.dict(sys.modules, {"pypdfium2": None}):
            with pytest.raises(DependencyError) as exc:
                extract_pdf_layout_per_mille(dummy_pdf_path)
            assert "pypdfium2" in str(exc.value)

    def test_pdfium_crash_cleanup(self, mock_dependencies, dummy_pdf_path):
        """Ensure file is closed even if processing crashes."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = Mock()
        mock_pdfium.PdfDocument.return_value = mock_doc

        # Crash during iteration
        mock_doc.__iter__ = Mock(side_effect=ValueError("Deep internal crash"))

        with pytest.raises(PDFProcessingError):
            extract_pdf_layout_per_mille(dummy_pdf_path)

        mock_doc.close.assert_called_once()


# --- Tests for extract_pdf_pages ---


class TestRenderPages:
    def test_render_all_pages(self, mock_dependencies, dummy_pdf_path):
        mock_pdfium, _ = mock_dependencies
        mock_doc = MagicMock()
        mock_page = Mock()
        mock_bitmap = Mock()

        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__len__.return_value = 1
        mock_doc.__getitem__.return_value = mock_page

        mock_page.render.return_value = mock_bitmap
        mock_bitmap.to_pil.return_value = "PILImage"

        gen = extract_pdf_pages(dummy_pdf_path)
        images = list(gen)

        assert len(images) == 1
        assert images[0] == "PILImage"
        mock_doc.close.assert_called_once()

    def test_render_skipped_invalid_index(self, mock_dependencies, dummy_pdf_path):
        """Test skipping out of bounds pages."""
        mock_pdfium, _ = mock_dependencies
        mock_doc = MagicMock()
        mock_pdfium.PdfDocument.return_value = mock_doc
        mock_doc.__len__.return_value = 5  # 0-4 valid

        # Request 10 (invalid)
        gen = extract_pdf_pages(dummy_pdf_path, pages=[10])
        images = list(gen)
        assert len(images) == 0


class TestOCRExtraction:
    @pytest.fixture
    def mock_pytesseract(self):
        """
        Creates a mock pytesseract module with necessary Exception classes attached.
        """
        mock = MagicMock()
        # Define the specific exceptions pytesseract uses so catches work
        mock.TesseractNotFoundError = type("TesseractNotFoundError", (Exception,), {})
        mock.TesseractError = type("TesseractError", (Exception,), {})
        return mock

    def test_ocr_missing_library(self, dummy_pdf_path):
        """Ensure DependencyError is raised if pytesseract pip package is missing."""
        with patch.dict(sys.modules, {"pytesseract": None}):
            with pytest.raises(DependencyError) as exc:
                ocr_extract_pdf_text(dummy_pdf_path)
            assert "pip install pytesseract" in str(exc.value)

    def test_ocr_missing_binary(self, dummy_pdf_path, mock_pytesseract):
        """Ensure DependencyError is raised if tesseract binary is not in PATH."""
        # 1. Setup the check to fail
        mock_pytesseract.get_tesseract_version.side_effect = mock_pytesseract.TesseractNotFoundError("Binary not found")

        with patch.dict(sys.modules, {"pytesseract": mock_pytesseract}):
            with pytest.raises(DependencyError) as exc:
                ocr_extract_pdf_text(dummy_pdf_path)

            # Verify we point users to the official docs, not a specific command
            assert "tesseract-ocr.github.io" in str(exc.value)

    @patch("dorsal.file.preprocessing.pdf.extract_pdf_pages")
    def test_ocr_happy_path(self, mock_extract_pages, dummy_pdf_path, mock_pytesseract):
        """Test successful text extraction from pages."""
        # 1. Setup Mock Images
        mock_image_1 = Mock(name="Image1")
        mock_image_2 = Mock(name="Image2")
        mock_extract_pages.return_value = iter([mock_image_1, mock_image_2])

        # 2. Setup OCR responses
        mock_pytesseract.image_to_string.side_effect = ["Page One Text", "Page Two Text"]

        with patch.dict(sys.modules, {"pytesseract": mock_pytesseract}):
            results = ocr_extract_pdf_text(dummy_pdf_path, language="fra", config="--psm 6", render_scale=4.0)

        # 3. Assertions
        assert len(results) == 2
        assert results[0] == "Page One Text"
        assert results[1] == "Page Two Text"

        # Verify arguments passed to renderer
        mock_extract_pages.assert_called_once_with(dummy_pdf_path, scale=4.0, password=None)

        # Verify arguments passed to tesseract
        mock_pytesseract.image_to_string.assert_has_calls(
            [call(mock_image_1, lang="fra", config="--psm 6"), call(mock_image_2, lang="fra", config="--psm 6")]
        )

    @patch("dorsal.file.preprocessing.pdf.extract_pdf_pages")
    def test_ocr_partial_failure(self, mock_extract_pages, dummy_pdf_path, mock_pytesseract):
        """Test that if one page fails OCR, we log it and continue (returning empty str), rather than crashing."""
        mock_extract_pages.return_value = iter([Mock(), Mock()])

        # First page fails, Second succeeds
        mock_pytesseract.image_to_string.side_effect = [
            mock_pytesseract.TesseractError(1, "Error processing image"),
            "Success Text",
        ]

        with patch.dict(sys.modules, {"pytesseract": mock_pytesseract}):
            results = ocr_extract_pdf_text(dummy_pdf_path)

        assert len(results) == 2
        assert results[0] == ""  # Empty string on failure
        assert results[1] == "Success Text"



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import re
from typing import Annotated, Any, Literal

from pydantic import AfterValidator, BaseModel, Field

logger = logging.getLogger(__name__)


def get_dataset_id(dataset_string: str | Any) -> str:
    """Parses dataset_id out of a string.

    Example: extracts 'namespace/name' from a potentially prefixed string
        'https://dorsalhub.com/d/dorsal/iso-language' -> 'dorsal/iso-language'

    Args:
        dataset_string: User-provided dataset string. Expected format: 'namespace/name'.

    Returns:
        str: The dataset ID.

    Raises:
        ValueError: Not a string or no Dataset ID found.

    """
    if not isinstance(dataset_string, str):
        logger.warning("Invalid type: expected str, got %s.", type(dataset_string).__name__)
        raise ValueError(f"Dataset ID must be a string, got {type(dataset_string).__name__}.")

    match = re.match(
        pattern=r"^(?:.*\/)?(?P<dataset_id>[a-z0-9\-]{3,32}\/[a-z0-9\-]{3,32})$",
        string=dataset_string,
    )

    if match:
        extracted_id = match.group("dataset_id")
        logger.debug(
            "Successfully extracted Dataset ID: '%s' from input string: '%s'.",
            extracted_id,
            dataset_string,
        )
        return extracted_id
    else:
        logger.warning(
            "Invalid Dataset ID format: '%s'. No valid ID pattern found.",
            dataset_string,
        )
        raise ValueError(f"Invalid Dataset ID format: '{dataset_string}'. Expected 'namespace/name'.")


def check_no_double_hyphens(value: str) -> str:
    """Raises ValueError if the string contains double-hyphens"""
    if "--" in value:
        raise ValueError("must not contain double hyphens")
    return value


NAMESPACE_ID_OR_DATASET_NAME_REGEX = r"^[a-z0-9\-]{3,32}$"

DATASET_ID_REGEX = r"^[a-z0-9\-]{3,32}\/[a-z0-9\-]{3,32}$"
RX_DATASET_ID = re.compile(DATASET_ID_REGEX)

DatasetNamespace = Annotated[str, Field(pattern=r"^[a-z0-9\-]{3,32}$")]
DatasetName = Annotated[str, Field(pattern=r"^[a-z0-9\-]{3,32}$")]
DatasetID = Annotated[str, Field(pattern=DATASET_ID_REGEX), AfterValidator(check_no_double_hyphens)]


def is_valid_dataset_id_or_schema_id(value: str | Any) -> bool:
    if not isinstance(value, str):
        return False
    return bool(RX_DATASET_ID.match(value))


class Dataset(BaseModel):
    dataset_id: DatasetID
    type: Literal["Reference", "File"]
    key_field: str
    version: str
    dataset_schema: dict = Field(alias="schema")
    date_created: datetime.datetime
    date_modified: datetime.datetime
